# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

"""Set up logging for mots."""
import logging
import logging.config
from pathlib import Path

from mots.settings import settings

logger = logging.getLogger(__name__)


def init_logging(
    debug: bool = False,
    path: Path = settings.LOG_FILE,
    log_max_size: int = settings.LOG_MAX_SIZE,
    log_backups: int = settings.LOG_BACKUPS,
):
    """Initialize default logger."""
    logging.config.dictConfig(
        {
            "version": 1,
            "formatters": {
                "standard": {
                    "format": "%(asctime)-13s %(module)-10s %(levelname)-8s %(message)s"
                }
            },
            "handlers": {
                "console": {
                    "class": "logging.StreamHandler",
                    "formatter": "standard",
                    "level": logging.DEBUG if debug else logging.INFO,
                },
                "file": {
                    "class": "logging.handlers.RotatingFileHandler",
                    "formatter": "standard",
                    "filename": path,
                    "maxBytes": log_max_size,
                    "backupCount": log_backups,
                    "encoding": "utf8",
                    "level": logging.DEBUG,
                },
            },
            "loggers": {
                "mots": {
                    "handlers": ["console", "file"],
                    "level": logging.DEBUG,
                },
            },
        }
    )
    logger.debug(f"Logging configured with log file @ {path}.")



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

"""Test config module."""

from mots.config import (
    calculate_hashes,
    FileConfig,
    reference_anchor_for_module,
)
from mots.directory import Directory


def test_calculate_hashes(config):
    export = b""
    hashes = calculate_hashes(config, export)[1]

    assert (
        hashes["config"] == "76997b8d70e561c7ec8a21e1cefbc3c658c6d91a"
    ), "Was `conftest.config` changed?"
    assert hashes["export"] == "da39a3ee5e6b4b0d3255bfef95601890afd80709"


def test_FileConfig__check_hashes(repo):
    file_config = FileConfig(repo / "mots.yml")
    file_config.load()
    errors = file_config.check_hashes()
    assert errors == []

    file_config.config["hashes"]["config"] = "asdf"
    file_config.config["hashes"]["export"] = "ghjk"
    errors = file_config.check_hashes()
    assert errors == [
        "Mismatch in config hash detected.",
        "76997b8d70e561c7ec8a21e1cefbc3c658c6d91a does not match asdf",
        "config file is out of date.",
        "Mismatch in export hash detected.",
        "da39a3ee5e6b4b0d3255bfef95601890afd80709 does not match ghjk",
        "export file is out of date.",
    ]


def test_reference_anchor_for_module(repo):
    """Test that a reference to an existing person is correctly updated.

    When a reference to a person references the same person but a different dictionary,
    reference_anchor_for_module should update that reference to reflect the existing
    person.
    """
    file_config = FileConfig(repo / "mots.yml")
    file_config.load()
    directory = Directory(file_config)
    directory.load(full_paths=True)

    module = file_config.config["modules"][0]["submodules"][0]
    owner_emeritus_field = module["meta"]["owners_emeritus"]
    # Check that existing owner emeritus references the person in the roster.
    assert len(owner_emeritus_field) == 1
    assert id(owner_emeritus_field[0]) == id(file_config.config["people"][2])

    # Change owner emeritus to be the first person in the roster.
    new_person = {"name": "jane", "nick": "jane", "bmo_id": 0}
    owner_emeritus_field[0] = new_person

    assert new_person == file_config.config["people"][0]
    assert id(owner_emeritus_field[0]) != id(file_config.config["people"][0])

    key = "owners_emeritus"
    index = 0
    reference_anchor_for_module(
        index, file_config.config["people"][0], key, file_config, directory, module
    )

    assert id(owner_emeritus_field[0]) == id(file_config.config["people"][0])



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

"""CI/CD related methods."""

import logging
import os

from mots import __version__
from mots.settings import settings

logger = logging.getLogger(__name__)


def validate_version_tag():
    """Check the current version against the CircleCI tag."""
    KEY = settings.CIRCLECI_TAG_KEY
    circle_tag = os.getenv(KEY)
    if not circle_tag:
        raise ValueError(f"{KEY} environment variable missing or blank.")

    if circle_tag != __version__:
        raise ValueError(
            f"{KEY} does not match version ({circle_tag} vs. {__version__})"
        )

    logger.info(f"{KEY} matches version.")



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

"""Tests for utils module."""

from unittest.mock import MagicMock

import pytest

from mots.utils import (
    generate_machine_readable_name,
    mkdir_if_not_exists,
    parse_real_name,
    touch_if_not_exists,
)


@pytest.fixture(scope="function")
def mock_path():
    def _mock_path(exists=False, is_dir=False, is_file=False):
        assert exists is True or exists is False
        if not exists:
            assert is_file is False and is_dir is False
        else:
            assert is_file is False or is_dir is False

        path = MagicMock()
        path.exists.return_value = exists
        if exists:
            path.is_dir.return_value = is_dir
            path.is_file.return_value = is_file
        else:
            path.is_dir.return_value = False
            path.is_file.return_value = False
        return path

    return _mock_path


def test_generate_machine_readable_name():
    assert generate_machine_readable_name("Hello, world!") == "hello_world"
    assert generate_machine_readable_name("test") == "test"
    assert generate_machine_readable_name("Another test !@%#@") == "another_test"
    assert generate_machine_readable_name("User Interface (UI)") == "user_interface_ui"
    assert generate_machine_readable_name("test: testing") == "test_testing"


def test_parse_real_name():
    assert parse_real_name("tëstér testerson (:test) [te/st]") == {
        "name": "tëstér testerson",
        "info": "(:test) [te/st]",
    }

    assert parse_real_name("tester testerson [:test] (te/st)") == {
        "name": "tester testerson",
        "info": "[:test] (te/st)",
    }

    assert parse_real_name("tester testerson (te/st)") == {
        "name": "tester testerson",
        "info": "(te/st)",
    }

    assert parse_real_name("tester testerson :test") == {
        "name": "tester testerson",
        "info": ":test",
    }

    # This test is not great, but some folks have their e-mail instead of name.
    assert parse_real_name("tester@testerson.example (:test)") == {
        "name": "tester",
        "info": "@testerson.example (:test)",
    }

    assert parse_real_name("Emoji Tester💡 :tester") == {
        "name": "Emoji Tester",
        "info": "💡 :tester",
    }

    assert parse_real_name("tester testerson | TEST | te/st") == {
        "name": "tester testerson",
        "info": "| TEST | te/st",
    }

    assert parse_real_name("tester testerson") == {
        "name": "tester testerson",
        "info": "",
    }

    assert parse_real_name("tester") == {
        "name": "tester",
        "info": "",
    }

    assert parse_real_name("(:bees)") == {
        "name": "",
        "info": "(:bees)",
    }

    assert parse_real_name("[:bees]") == {
        "name": "",
        "info": "[:bees]",
    }

    assert parse_real_name("[bees]") == {
        "name": "",
        "info": "[bees]",
    }

    assert parse_real_name("(bees)") == {
        "name": "",
        "info": "(bees)",
    }

    assert parse_real_name("") == {
        "name": "",
        "info": "",
    }


def test_mkdir_if_not_exists__does_not_exist(mock_path):
    path = mock_path(exists=False)
    mkdir_if_not_exists(path)
    assert path.exists.call_count == 1
    assert path.mkdir.call_count == 1


def test_mkdir_if_not_exists__exists_as_file(mock_path):
    path = mock_path(exists=True, is_file=True)
    mkdir_if_not_exists(path)
    assert path.exists.call_count == 2
    assert path.mkdir.call_count == 0


def test_mkdir_if_not_exists__exists(mock_path):
    path = mock_path(exists=True, is_dir=True)
    mkdir_if_not_exists(path)
    assert path.exists.call_count == 2
    assert path.mkdir.call_count == 0


def test_touch_if_not_exists__does_not_exist(mock_path):
    path = mock_path(exists=False)
    touch_if_not_exists(path)
    assert path.exists.call_count == 1
    assert path.touch.call_count == 1


def test_touch_if_not_exists__exists_as_dir(mock_path):
    path = mock_path(exists=True, is_dir=True)
    touch_if_not_exists(path)
    assert path.exists.call_count == 2
    assert path.touch.call_count == 0


def test_touch_if_not_exists__exists(mock_path):
    path = mock_path(exists=True, is_file=True)
    touch_if_not_exists(path)
    assert path.exists.call_count == 2
    assert path.touch.call_count == 0



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

"""Set up the mots environment.

This module fetches various variables either from the environment or config files that
will directly affect the operation of mots after startup.

Variables defined in this module may be overridden by environment variables or
in the overrides file.
"""

import os
from pathlib import Path

from mots import __version__
from mots.yaml import yaml

HOME_DIRECTORY = Path.home()
RESOURCE_DIRECTORY = HOME_DIRECTORY / ".mots"
OVERRIDES_FILE = RESOURCE_DIRECTORY / "settings.yaml"


class Settings:
    """Helper class to convert settings dict to object."""

    DEFAULTS = {
        "BUGZILLA_API_KEY": "",
        "BUGZILLA_URL": "https://bugzilla.mozilla.org/rest",
        "CHECK_DEV_RELEASES": 0,
        "CHECK_FOR_UPDATES": 1,
        "CIRCLECI_TAG_KEY": "CIRCLE_TAG",
        "DEBUG": 0,
        "DEFAULT_CONFIG_FILEPATH": Path("./mots.yaml"),
        "DEFAULT_EXPORT_FORMAT": "rst",
        "LOG_BACKUPS": 5,
        "LOG_FILE": RESOURCE_DIRECTORY / "mots.log",
        "LOG_MAX_SIZE": 1024 * 1024 * 50,
        "OVERRIDES_FILE": OVERRIDES_FILE,
        "PMO_COOKIE": "",
        "PMO_SEARCH_URL": "https://people.mozilla.org/s?query=",
        "PMO_URL": "https://people.mozilla.org/api/v4",
        "RESOURCE_DIRECTORY": RESOURCE_DIRECTORY,
        "SEARCHFOX_BASE_URL": "https://searchfox.org",
        "USER_AGENT": f"mots/{__version__}",
        "VERSION": __version__,
    }

    # NOTE: keys present in SECURE_KEYS will not be added to logs. Users will be
    # prompted using `getpass`, however they will still be stored as cleartext
    # in the settings file, like all other values.
    SECURE_KEYS = ("BUGZILLA_API_KEY",)

    def __init__(self, **kwargs):
        self.settings = {}
        self.overrides = kwargs
        self.load()
        self.set_attributes()

    def load(self):
        """Load settings from environment where possible, or use defaults."""
        # First get type from default, then load values from env if possible.
        self.settings = {
            k: type(v)(os.getenv(f"MOTS_{k}", v)) for k, v in self.DEFAULTS.items()
        }

        # Now, update with override values if possible.
        self.settings.update(self.overrides)

    def set_attributes(self):
        """Set attributes on instance using settings dict."""
        for k, v in self.settings.items():
            setattr(self, k, v)


if OVERRIDES_FILE.exists():
    with OVERRIDES_FILE.open("r") as f:
        overrides = yaml.load(f) or {}
else:
    overrides = {}
settings = Settings(**overrides)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

"""This module sets up parsers and maps cli commands to methods."""

from datetime import datetime
from pathlib import Path
import argparse
import getpass
import logging
import sys

from mots import module
from mots import config
from mots.bmo import MissingBugzillaAPIKey, BMOClient
from mots.ci import validate_version_tag
from mots.directory import Directory
from mots.export import export_to_format
from mots.logging import init_logging
from mots.settings import settings
from mots.utils import (
    get_list_input,
    mkdir_if_not_exists,
    touch_if_not_exists,
    check_for_updates as _check_for_updates,
)
from mots.yaml import yaml
from mots import __version__


logger = logging.getLogger(__name__)


def init(args: argparse.Namespace) -> None:
    """Initialize mots configuration file."""
    file_config = config.FileConfig(Path(args.path))
    file_config.init()


def ls(args: argparse.Namespace) -> None:
    """List modules."""
    file_config = config.FileConfig(Path(args.path))
    file_config.load()
    module.ls(file_config.config["modules"])


def show(args: argparse.Namespace) -> None:
    """Show given module details."""
    file_config = config.FileConfig(Path(args.path))
    file_config.load()
    module.show(file_config.config["modules"], args.module)


def validate(args: argparse.Namespace) -> None:
    """Validate configuration and show error output if applicable."""
    file_config = config.FileConfig(Path(args.path))
    file_config.load()
    errors = config.validate(file_config.config, file_config.repo_path) or []
    for error in errors:
        logger.error(error)
    sys.exit(1 if errors else 0)


def clean(args: argparse.Namespace) -> None:
    """Run clean methods for configuration file and write to disk."""
    file_config = config.FileConfig(Path(args.path))
    file_config.load()
    try:
        config.clean(file_config)
    except MissingBugzillaAPIKey:
        logger.error("Could not detect a Bugzilla API Key.")
        messages = (
            "Either set it in your environment (MOTS_BUGZILLA_API_KEY) or add it to ",
            "your settings file by running: ",
            "`mots settings write BUGZILLA_API_KEY`.",
            "You will be prompted to enter the API key after running this command.",
            "You can generate a Bugzilla API key in your User Preferences page under",
            "the API Keys tab.",
        )
        print("\n".join(messages))
        sys.exit(1)


def check_hashes(args: argparse.Namespace) -> None:
    """Check stored hashes against calculated hashes and exit with appropriate code."""
    file_config = config.FileConfig(Path(args.path))
    file_config.load()
    errors = file_config.check_hashes()
    if errors:
        for error in errors:
            logger.error(error)
        sys.exit(1)


def query(args: argparse.Namespace) -> None:
    """Query list of files for module information."""
    file_config = config.FileConfig(Path(args.path))
    file_config.load()
    directory = Directory(file_config)
    directory.load()
    result = directory.query(*args.paths)
    for path, modules in result.path_map.items():
        module_names = ",".join([m.machine_name for m in modules])
        print(f"{path}:{module_names}\n")


def add(args: argparse.Namespace) -> None:
    """Add a new module or submodule to repo configuration file."""
    params = {
        "machine_name": input(
            "Enter a machine name (alphanumeric characters and underscores)"
            " for the new module (e.g. core_accessibility): "
        ),
        "name": input("Enter a human readable name (e.g. Core: Accessibility): "),
        "description": input("Enter a description for the new module: "),
        "owners": get_list_input("Enter a comma separated list of owner bugzilla IDs"),
        "peers": get_list_input("Enter a comma separated list of peer bugzilla IDs"),
        "includes": get_list_input("Enter a comma separated list of paths to include"),
        "excludes": get_list_input("Enter a comma separated list of paths to exclude"),
    }

    params["owners"] = [{"bmo_id": int(bmo_id)} for bmo_id in params["owners"]]
    params["peers"] = [{"bmo_id": int(bmo_id)} for bmo_id in params["peers"]]

    parent = input("Enter a machine name of the parent module (optional): ") or None
    file_config = config.FileConfig(Path(args.path))
    config.add(params, file_config, parent=parent, write=True)


def ci(args: argparse.Namespace) -> None:
    """Perform CI checks or validations."""
    try:
        validate_version_tag()
    except ValueError as e:
        logger.error(e)
        sys.exit(1)


def version():
    """Return version information."""
    return __version__


def export(args: argparse.Namespace) -> None:
    """Export repo configuration and write to disk, or print to stdout as needed."""
    file_config = config.FileConfig(Path(args.path))
    file_config.load()
    directory = Directory(file_config)
    directory.load()

    frmt = (
        args.format
        if args.format
        else file_config.config.get("export", {}).get(
            "format", settings.DEFAULT_EXPORT_FORMAT
        )
    )

    # Render output based on provided format.
    output = export_to_format(directory, frmt)

    if not args.out:
        # Explicit output path was not provided, try to get it from config.
        if file_config.config.get("export", {}).get("path"):
            out_path = Path(file_config.config["export"]["path"])
            with out_path.open("w", encoding="utf-8") as f:
                logger.info(f"Writing output to specified file path ({out_path})...")
                f.write(output)
        else:
            # No output path could be determined, so output to standard out.
            print(output)
    else:
        # TODO: do more checks here to make sure we don't overwrite important things.
        logger.info(f"Writing output to specified file path ({args.out})...")
        with args.out.open("w", encoding="utf-8") as f:
            f.write(output)


def export_and_clean(args: argparse.Namespace) -> None:
    """Run clean, export, and clean again to save users keystrokes."""
    # The first clean checks mots.yaml for any issues and synchronizes with BMO.
    clean(args)

    # The export exports the yaml file to rst. This assumes that the export path is
    # defined in the configuration.
    export(args)

    # The last clean resets hashes
    clean(args)


def write(args: argparse.Namespace):
    """Set a specified settings variable to the provided value."""
    key = args.key[0]

    with settings.OVERRIDES_FILE.open("r", encoding="utf-8") as f:
        overrides = yaml.load(f) or {}

    if key not in settings.DEFAULTS:
        raise ValueError(f"{key} does not exist in defaults.")

    is_secure = key in settings.SECURE_KEYS

    if args.value is not None:
        value = args.value
    else:
        prompt = f"Enter value for {key}: "
        value = input(prompt) if not is_secure else getpass.getpass(prompt=prompt)

    _type = type(settings.DEFAULTS[key])

    if _type(value) == overrides.get(key):
        raise ValueError(f"{key} is already set to {value} in overrides file.")

    overrides[key] = _type(value)
    value_output = value if not is_secure else "*" * 10
    logger.info(
        f"{key} is now set to {value_output} ({_type.__name__}) in overrides file."
    )

    with settings.OVERRIDES_FILE.open("w", encoding="utf-8") as f:
        yaml.dump(overrides, f)


def read(args: argparse.Namespace):
    """Print the value of a specified settings variable.

    If no key is provided, prints all available settings variables.
    """

    def out_template(key: str, value: str):
        return f"{key}: {value} ({type(value).__name__})"

    if not args.key:
        for key, value in settings.settings.items():
            print(out_template(key, value))
        return

    key = args.key

    if key not in settings.settings:
        raise ValueError(f"{key} does not exist in defaults.")
    value = settings.settings[key]
    print(out_template(key, value))


def search(args: argparse.Namespace):
    """Search Bugzilla API for users, given an email address."""
    bmo_client = BMOClient()
    result = bmo_client.get_matches(args.match)
    logger.info(f"Found {len(result)} users.")
    logger.info("Tip: you can copy and paste a dictionary entry into a field.")

    for user in result:
        person = {"bmo_id": user["id"], "name": user["real_name"], "nick": user["nick"]}
        out = f"{person} -> {user['email']}"
        print(out)


def check_for_updates(args: argparse.Namespace) -> None:
    """CLI wrapper around _check_for_udpdates."""
    _check_for_updates(args.include_dev_releases)


def call_main_with_args():
    """Call `main` after creating a parser and parsing arguments.

    This method is used when running mots directly via the CLI. When mots is run via
    an integration (e.g., mach), then the integration must call `main` directly with
    the arguments.
    """
    parser = create_parser()
    args = parser.parse_args()
    main(args)


def main(
    args: argparse.Namespace,
    skip_update_check: bool = False,
) -> None:
    """Run startup commands and redirect to appropriate function."""
    mkdir_if_not_exists(settings.RESOURCE_DIRECTORY)
    touch_if_not_exists(settings.OVERRIDES_FILE)
    init_logging(debug=args.debug)

    if hasattr(args, "func"):
        if skip_update_check:
            logger.debug("Skipping update check.")
        elif args.func != check_for_updates and settings.CHECK_FOR_UPDATES:
            try:
                _check_for_updates(settings.CHECK_DEV_RELEASES)
            except Exception as e:
                logger.warning("Could not check for updates.")
                logger.debug(e)
        logger.debug(f"Calling {args.func} with {args}...")
        st = datetime.now()
        try:
            args.func(args)
        except Exception as e:
            logger.exception(e)
        else:
            logger.info("Success!")
        et = datetime.now()
        logger.debug(f"{args.func} took {(et - st).total_seconds()} seconds.")
    else:
        # By default, print help to screen.
        create_parser().print_help()


def _add_path_argument(_parser):
    _parser.add_argument(
        "--path",
        "-p",
        type=Path,
        help="the path of the repo config file",
        default=settings.DEFAULT_CONFIG_FILEPATH,
    )


def create_parser(subcommand=None):
    """Create parser, subparsers, and arguments."""
    parsers = {}
    parser = argparse.ArgumentParser(description="main command line interface for mots")
    parser.add_argument("--debug", action="store_true", help="enable debug output")
    parser.add_argument("--version", action="version", version=version())
    main_cli = parser.add_subparsers(title="commands")
    module_parser = main_cli.add_parser("module", help="module operations")
    _add_path_argument(module_parser)
    module_cli = module_parser.add_subparsers(title="module")

    settings_parser = main_cli.add_parser("settings", help="settings operations")
    settings_cli = settings_parser.add_subparsers(title="settings")

    user_parser = main_cli.add_parser("user", help="user operations")
    user_cli = user_parser.add_subparsers(title="user")

    # Create path argument template.
    path_flags = ("--path", "-p")
    path_args = {
        "type": Path,
        "help": "the path of the repo config file",
        "default": settings.DEFAULT_CONFIG_FILEPATH,
    }

    for _cli, func, help_text in (
        (main_cli, ci, "perform CI checks or operations"),
        (main_cli, init, "initialize mots configuration in repo"),
        (main_cli, clean, "clean mots config"),
        (main_cli, check_hashes, "check mots config and export hashes"),
        (main_cli, query, "query the module directory"),
        (main_cli, export, "export the module directory"),
        (main_cli, export_and_clean, "perform automatic cleaning and exporting"),
        (main_cli, validate, "validate mots config"),
        (main_cli, check_for_updates, "check for new versions of mots"),
        (module_cli, add, "add a new module"),
        (module_cli, ls, "list all modules"),
        (module_cli, show, "show module details"),
        (settings_cli, write, "update settings variable and save to disk"),
        (settings_cli, read, "get settings variable, or all if no key is provided"),
        (user_cli, search, "search Bugzilla user database"),
    ):
        name = func.__name__.replace("_", "-")
        parsers[name] = _cli.add_parser(name, help=help_text)
        parsers[name].set_defaults(func=func)

    # Add custom arguments where needed.
    parsers["query"].add_argument("paths", nargs="+", help="a list of paths to query")
    parsers["query"].add_argument(*path_flags, **path_args)

    parsers["export"].add_argument(
        "--format", "-f", type=str, choices=("rst",), help="the format of exported data"
    )
    parsers["export"].add_argument(
        "--out", "-o", type=Path, help="the file path to output to"
    )
    parsers["export"].add_argument(*path_flags, **path_args)
    parsers["export-and-clean"].add_argument(*path_flags, **path_args)
    parsers["export-and-clean"].add_argument(
        "--format", "-f", type=str, choices=("rst",), help="the format of exported data"
    )
    parsers["export-and-clean"].add_argument(
        "--out", "-o", type=Path, help="the file path to output to"
    )

    parsers["write"].add_argument("key", nargs=1, help="the settings key to set")
    parsers["write"].add_argument(
        "value", nargs="?", help="the value to set key to if present"
    )

    parsers["read"].add_argument(
        "key",
        nargs="?",
        help="fetch the value of key if provided, or all keys otherwise",
    )

    parsers["search"].add_argument("match", nargs=1, help="a search string")
    parsers["check-for-updates"].add_argument(
        "--include-dev-releases",
        "-d",
        action="store_true",
        help="include dev releases in check",
        default=settings.CHECK_DEV_RELEASES,
    )

    parsers["init"].add_argument(*path_flags, **path_args)
    parsers["validate"].add_argument(*path_flags, **path_args)
    parsers["clean"].add_argument(*path_flags, **path_args)
    parsers["check-hashes"].add_argument(*path_flags, **path_args)

    if not subcommand:
        return parser
    if subcommand not in parsers:
        raise ValueError(f"{subcommand} not found.")
    return parsers[subcommand]



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

"""Module to load YAML parser."""

import logging

from ruamel.yaml import YAML

logger = logging.getLogger(__name__)


def load_yaml() -> YAML:
    """Load and return a ruamel.yaml.YAML instance."""
    yaml = YAML()
    yaml.indent(
        mapping=2,
        sequence=4,
        offset=2,
    )
    return yaml


yaml = load_yaml()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

"""
Mots is the in-tree module ownership system.

This library helps you track who owns what parts of the source code!
"""

from sys import version_info

if version_info >= (3, 8):
    from importlib.metadata import version
else:
    from importlib_metadata import version

__version__ = version("mots")



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

"""Test export functionalities."""

from unittest import mock

import pytest

from mots.export import (
    export_to_format,
    escape_for_rst,
    format_paths_for_rst,
    format_people_for_rst,
    format_emeritus,
)


@mock.patch("mots.export.Exporter")
def test_export_to_format(exporter):
    directory = mock.MagicMock()
    test = export_to_format(directory, frmt="rst")
    assert test == exporter(directory)._export_to_rst()

    with pytest.raises(ValueError):
        export_to_format(directory, frmt="unsupported-format")


def test_escape_for_rst():
    test_string = (
        "A couple of backslashes \\ \\\n"
        "A couple of asterisks * *\n"
        "A couple of backticks ` `\n"
        "All on one line: \\ * `"
    )

    expected_string = (
        "A couple of backslashes \\\\ \\\\\n"
        "A couple of asterisks \\* \\*\n"
        "A couple of backticks \\` \\`\n"
        "All on one line: \\\\ \\* \\`"
    )

    assert expected_string == escape_for_rst(test_string)


def test_export_format_paths_for_rst():
    """Ensure outputted strings are correct when formatting paths."""
    mock_directory = mock.MagicMock()
    mock_directory.config_handle.config = {
        "export": {"searchfox_enabled": True},
        "repo": "test-repo",
    }

    paths = ["test_path_1", "test_path_2"]

    assert format_paths_for_rst(paths, indent=8, directory=mock_directory) == (
        "\n        | `test_path_1 "
        "<https://searchfox.org/test-repo/search?q=&path=test_path_1>`__"
        "\n        | `test_path_2 "
        "<https://searchfox.org/test-repo/search?q=&path=test_path_2>`__"
    )

    mock_directory.config_handle.config = {
        "repo": "test-repo",
    }

    paths = ["test_path_1", "test_path_2"]

    assert format_paths_for_rst(paths, indent=8, directory=mock_directory) == (
        "\n        | test_path_1" "\n        | test_path_2"
    )


def test_export_format_people_for_rst(config):
    """Ensure outputted strings are correct when formatting people."""
    config["people"].append({"nick": "unnamed"})
    test = format_people_for_rst(config["people"], indent=8)
    assert test == (
        "\n        | `jane (jane) <https://people.mozilla.org/s?query=jane>`__"
        "\n        | `jill (jill) <https://people.mozilla.org/s?query=jill>`__"
        "\n        | `otis (otis) <https://people.mozilla.org/s?query=otis>`__"
        "\n        | `unnamed <https://people.mozilla.org/s?query=unnamed>`__"
    )


def test_export_format_emeritus(config):
    emeritus = ["maggie", config["people"][0]]
    test = format_emeritus(emeritus)
    assert test == "maggie, jane"



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

"""Module operations and utility functions."""

from __future__ import annotations

import logging
from pathlib import Path
from pprint import pprint as print

logger = logging.getLogger(__name__)


class Module:
    """A top-level module or a submodule.

    :param machine_name: a unique, machine-readable name for the module
    :param repo_path: the path of the top-level repository
    :param name: the name of the module
    :param description: the description of the module
    :param includes: a list of paths (glob format) to include
    :param excludes: a list of paths (glob format) to exclude
    :param owners: a list of owners that will own all paths in this module
    :param peers: a list of peers for this module
    :param meta: a dictionary of meta data related to this module
    :param parent: the parent module of this module, if this is a submodule
    :param submodules: a list of submodules of this module
    :param exclude_submodule_paths: when True, paths in submodules are excluded
    :param exclude_module_paths: when True, common paths in other modules are excluded
        from the directory index

    The ``Module`` class wraps modules and submodules in a configuration, and provides
    useful methods for validation, calculating paths, and serialization.
    """

    def __repr__(self):
        """Return a human readable name for the instance."""
        return f"<Module: {self.machine_name}>"

    def __init__(
        self,
        machine_name: str,
        repo_path: str | Path,
        name: str = "",
        description: str = "",
        includes: list[str] | None = None,
        excludes: list[str] | None = None,
        owners: list[dict] | None = None,
        peers: list[dict] | None = None,
        meta: dict = None,
        parent: "Module" = None,
        submodules: list[dict] | None = None,
        exclude_submodule_paths: bool = True,
        exclude_module_paths: bool = False,
    ):
        self.name = name
        self.machine_name = machine_name
        self.description = description
        self.repo_path = Path(repo_path)
        self.parent = parent

        self.excludes = excludes or []
        self.includes = includes or []
        self.owners = owners or []
        self.peers = peers or []
        self.submodules = []
        self.exclude_submodule_paths = exclude_submodule_paths
        self.exclude_module_paths = exclude_module_paths
        self.errors = []

        if submodules:
            for module in submodules:
                self.submodules.append(
                    Module(parent=self, repo_path=repo_path, **module)
                )

        self.meta = meta

        # `includes`, `excludes`, and `owners` can be omitted in submodules, in which
        # case they will be inherited from their parent module.
        if not self.includes and self.parent:
            self.includes = self.parent.includes

        if not self.excludes and self.parent:
            self.excludes = self.parent.excludes

        if not self.owners and self.parent:
            self.owners = self.parent.owners

        if not self.peers and self.parent:
            self.peers = self.parent.peers

        self.owner_names = [owner.get("nick", "") for owner in self.owners]
        self.peer_names = [peer.get("nick", "") for peer in self.peers]

    def calculate_paths(self):
        """Calculate paths based on inclusions and exclusions.

        Upon calling this method, excluded paths are parsed using ``pathlib.Path.rglob``
        and then subtracted from the included paths, which are parsed in the same way.

        :rtype: set
        """
        includes = []
        for pattern in self.includes:
            logger.debug(f"Expanding {pattern} in {self.machine_name}...")
            expanded = list(self.repo_path.glob(pattern))
            logger.debug(
                f"Pattern {pattern} expanded to {len(expanded)} included path(s)."
            )
            includes += expanded

        excludes = []
        for pattern in self.excludes:
            logger.debug(f"Expanding {pattern} in {self.machine_name}...")
            expanded = list(self.repo_path.glob(pattern))
            logger.debug(
                f"Pattern {pattern} expanded to {len(expanded)} excluded path(s)."
            )
            excludes += expanded

        paths = set(includes) - set(excludes)
        if self.exclude_submodule_paths:
            for submodule in self.submodules:
                paths -= submodule.calculate_paths()
        return paths

    def serialize(self):
        """Return a dictionary with relevant module information.

        :rtype: dict
        """
        serialized = {
            "machine_name": self.machine_name,
            "name": self.name,
            "description": self.description,
            "includes": self.includes,
            "excludes": self.excludes,
            "owners": self.owners,
            "peers": self.peers,
            "meta": self.meta,
        }

        if self.submodules:
            serialized["submodules"] = [sm.serialize() for sm in self.submodules]
        if self.parent:
            serialized["parent"] = self.parent.machine_name

        return serialized

    def validate(self):
        """Perform validation on module and submodules recursively.

        Starting with the current module, ensure that this module includes at least one
        valid path and a valid name and machine name,  and then run the same validation
        on all submodules.

        :param errors: a list of errors to append to
        :rtype: list
        """
        errors = []

        if not self.machine_name.strip():
            errors.append("Module has a blank machine_name.")

        if " " in self.machine_name:
            errors.append(f"Machine name {self.machine_name} contains white space.")

        if not self.calculate_paths():
            errors.append(f"No valid paths were found in {self.machine_name}.")

        if self.submodules:
            for submodule in self.submodules:
                errors += submodule.validate()

        return errors


def ls(modules: list[Module]):
    """Print a list of given modules.

    :param modules: a list of :class:`Module` instances
    """
    print(modules)


def show(modules: list[Module], module: str):
    """Show details for a particular module.

    :param modules: a list of :class:`Module` instances
    :param module: the `machine_name` of a given :class:`Module`

    :raises ValueError: when no module matches the given machine name
    """
    for _module in modules:
        if _module["machine_name"] == module:
            print(_module)
            return
    raise ValueError(f"{module} not found.")
